#' @name chullpeel
#' @title Convex peeling
#' @description Approximate a set of 2-dimensional credible regions by
#' using convex peeling.

#' @param X a data.frame or matrix with two columns
#' @param qq the quantiles required
#'
#' @return A list of sets of row indices, one for each quantile in
#' \code{qq}, showing the points on the boundary of the approximate
#' credible region.  See the example for how to use these in plotting.
#'
#' @export
#' @examples
#' foo <- matrix(rnorm(400), ncol = 2)
#' foo <- foo %*% matrix(c(1, 0, 1, 1), 2, 2)
#' cpeel <- chullpeel(foo, c(0.95, 0.75, 0.50))
#' 
#' plot(foo, type = "p", pch = 16, col = "darkgrey", asp = 1)
#' for (pts in cpeel)
#'   polygon(foo[pts, , drop=FALSE])

chullpeel <- function(X, qq = c(0.99, 0.9, 0.75, 0.5, 0.25, 0.1)) {

    if (is.data.frame(X))
        X <- as.matrix(X)
    stopifnot(is.matrix(X), ncol(X) == 2, qq > 0, qq <= 1)
    if (!all(diff(qq) < 0)) {
        warning("Order of quantiles changed")
        qq <- rev(sort(qq))
    }

    n <- nrow(X)
    curr <- 1L:n
    ch <- c() # NULL
    robj <- vector("list", length(qq))
    
    for (i in seq(along = robj)) {
        target <- qq[i]
        while (length(curr) > target * n) {
            curr <- setdiff(curr, ch)
            ch <- curr[chull(X[curr, ])]
        }
        robj[[i]] <- ch
    }
    names(robj) <- paste(100 * qq, "%", sep = "")
    robj
}
