#' Create a level plot
#'
#' A level plot is a 2D filled contour plot.  \code{levelplot} uses
#' \code{\link{.filled.contour}} to create a colour field, and then
#' overlays contours on it using \code{\link{contour}}.  It does not
#' do any annotations, which can be added afterwards: see Examples.
#'
#' @param x Vector of x-ordinates, length \code{m}
#' @param y Vector of y-ordinates, length \code{n}
#' @param z Matrix of z-values, dimension \code{m} by \code{n}
#' @param xlim Range of x-ordinates
#' @param ylim Range of y-ordinates
#' @param zlim Range for colours (white outside this range)
#' @param asp Aspect ratio
#' @param levels Contour values
#' @param nlevels Number of levels (approximately)
#' @param col Colours, or colour palette
#' @param ... Additional arguments to \code{\link{contour}}
#'
#' @seealso \code{\link{filled.contour}}, which uses a separate colour
#' scale rather than superimposing contours with value label.
#'
#' @export
#' @examples
#' x <- seq(from = -2, to = 5, length = 101)
#' y <- seq(from = -2, to = 5, length = 101)
#' grid <- expand.grid(x = x, y = y)
#' z <- apply(as.matrix(grid), 1, banana, theta = 100)
#' dim(z) <- c(length(x), length(y))
#'
#' foo <- levelplot(x, y, log(z), asp = 1, lwd = 2, labcex = 0.7)
#'
#' ## now anotate
#'
#' title(xlab = "X values", ylab = "Y values",
#'       main = "Rosenbrock banana function")
#' axis(1, pos = foo$ylim[1])
#' axis(2, pos = foo$xlim[1], las = 1)

levelplot <- function(x, y, z,
                      xlim = range(x, finite = TRUE),
                      ylim = range(y, finite = TRUE),
                      zlim = range(z, finite = TRUE), asp = NA,
                      levels = pretty(zlim, nlevels),
                      nlevels = 7, col = cm.colors, ...) {

    stopifnot(is.matrix(z),
              is.vector(x, mode = "numeric"), length(x) == nrow(z),
              is.vector(y, mode = "numeric"), length(y) == ncol(z))
    plot.new()
    plot.window(xlim = xlim, ylim = ylim, asp = asp)
    if (is.function(col))
        col <- col(length(levels) - 1)
    else stopifnot(length(col) == length(levels) - 1)
    .filled.contour(x, y, z, levels, col)
    contour(x, y, z, levels = levels, ..., add = TRUE)
    invisible(list(xlim = xlim, ylim = ylim, asp = asp,
                   levels = levels, col = col))
}
