#' Create new design points similar to those supplied
#'
#' Take a set of points, and generate additional points that are close
#' to an identified subset of them.
#'
#' @param X design matrix
#' @param keep Subscript showing which rows of \code{X} to match with
#' the new points
#' @param n Number of additional points required
#' @param fac Expansion factor for over-sampling, see Details
#'
#' @return A new design matrix with \code{n} rows.  If \code{X} has
#' class \code{boxdesign}, then so does the new design.
#'
#' @details
#'
#' This is a stochastic method which generates new points in the same
#' hyper-rectangle as \code{X}, and then keeps the best \code{n},
#' according to the following rules:
#' 
#' \enumerate{
#' 
#' \item Only points that are nearer to a point in \code{X[keep, ]}
#' than they are to a point in \code{X[-keep, ]} are retained.
#'
#' \item If fewer than \code{n} points are retained, then the function
#' iterates until there are at least \code{n} points.
#' 
#' \item If there are more than \code{n} points retained, then each
#' point \code{z} is ranked in terms of its distance to the nearest
#' point in the union of \code{X[keep, ]} and \code{Z[-z, ]}, and the
#' \code{n} points with the largest distances are retained.
#' }
#'
#' \eqn{fac * n} points are generated in each iteration.  Try
#' increasing \code{fac} if the function seems slow.
#'
#' All interpoint distances are computed after standardisation.
#' \code{morepoints} has special handling for design matrices of class
#' \code{boxdesign}, for which the lower and upper limits are
#' available as an attribute.
#'
#' @seealso \code{\link{lhcube}} for generating space-filling designs
#' in the unit cube, and \code{\link{boxdesign}} for simple
#' rectangular designs.
#'
#' @export
#' @examples
#' X <- lhcube(101, 2); colnames(X) <- c("x", "y")
#' z <- apply(X, 1, banana, theta = 100)
#' plot(X, col = 1 + (z > 10), pch = 20)
#' moreX <- morepoints(X, keep = z > 10, n = 51)
#' points(moreX, col = "blue", pch = 20)

morepoints <- function(X, keep, n, fac = 10) {

    stopifnot(is.matrix(X))
    p <- ncol(X)
    rnx <- rownames(X)
    if (is.null(rnx)) {
        rnx <- funames(nrow(X))
        rownames(X) <- rnx
    }
    keep <- logisub(keep, rnx)
    if (!any(keep)) return(NULL)
    
    ## map to unit cube.  Can use the info in param for designs of
    ## class boxdesign.  Otherwise just use the empirical limits.

    nms <- colnames(X)
    if (is.null(nms)) {
        nms <- paste0("X", 1L:p)
        colnames(X) <- nms
    }

    if (inherits(X, "boxdesign")) {
        param <- attr(X, "param")
    } else {
        param <- tapply(X, col(X), range)
        names(param) <- nms
    }
    U <- tozeroone(X, param = param) # always take log. = NULL

    ## Z will contain the candidate new points

    Z <- matrix(numeric(), 0, p)

    while (nrow(Z) < n) {
        cand <- lhcube(ceiling(fac * n), p)
        dd <- mydist(U, cand)
        dbad <- apply(dd[!keep, , drop = FALSE], 2L, min)
        dgood <- apply(dd[keep, , drop = FALSE], 2L, min)
        ok <- dgood < dbad
        Z <- rbind(Z, cand[ok, , drop = FALSE])
    }

    ## now keep the best n

    if (nrow(Z) > n) {
        close <- mydist(rbind(Z, U[keep, , drop = FALSE]), Z)
        diag(close) <- Inf # remove self-distances of zero
        close <- apply(close, 2L, min)
        ok <- tail(order(close), n)
        Z <- Z[ok, , drop = FALSE]
    }

    ## back to the original ranges

    ranges <- do.call("rbind", param)
    Y <- sapply(1L:p, function(j) qunif(Z[, j], ranges[j, 1], ranges[j, 2]))
    dimnames(Y) <- list(funames(n, avoid = rnx), nms)

    if (inherits(X, "boxdesign")) {
        attr(Y, "param") <- param
        class(Y) <- "boxdesign"
    }
    Y
}

#### little internal function to process subscripts into a logical vector

logisub <- function(sub, x) {
    if (is.null(x)) stop("need non-NULL \'x\'")
    n <- length(x)
    if (is.logical(sub))
        robj <- rep(sub, length.out = n)
    else {
        robj <- rep(FALSE, n)
        if (is.numeric(sub) && all(sub %in% 1L:n))
            robj[sub] <- TRUE
        else if (is.character(sub) && all(sub %in% x))
            robj[match(sub, x)] <- TRUE
    }
    if (!any(robj)) stop("improper subscript")
    robj
}
