#' Principal variables
#'
#' Identify the principal variables in a matrix or dataframe.
#'
#' @param X Matrix or dataframe, must be numeric
#' @param as.names Return ordering as names or indices?
#'
#' @return A vector of names or indices, with a \code{scree} attribute
#' containing a vector of scree values (like a scree plot in principal
#' components)
#'
#' @details I apply the Cumming and Wooff (2007) algorithm to the
#' maximal linearly independent set of columns of \code{X}, which is
#' found using the \code{\link{ref}} function.
#'
#' @references
#' J.A. Cumming and D.A. Wooff, 2007, Dimension reduction via
#' principal variables, Computational Statistics and Data Analysis,
#' 52, 560-565.
#'
#' @export
#' @examples
#' pv <- prvar(state.x77, as.names = TRUE)
#' scree <- attr(pv, "scree")
#' plot(scree, ylim = c(0, 1), pch = 20,
#'   xlab = "", ylab = "Proportion of variation",
#'   main = "Principal variables scree plot",
#'   axes = FALSE, bty = "n")
#' axis(1, seq(along = scree), names(scree), las = 2, pos = 0, tick = FALSE)
#' axis(2)
#' abline(h = c(0, 1), col = "darkgrey")

prvar <- function(X, as.names = FALSE) {

  X <- as.matrix(X)
  p <- ncol(X)
  orignms <- colnames(X) 
  if (is.null(orignms))
    colnames(X) <- orignms <- paste("V", 1:p, sep = "")

  ## will need to compute the trace
  
  trace <- function(M) {
    M <- as.matrix(M)
    if (nrow(M) == 0) 0 else sum(diag(M))
  }

  ## columns with no variation are not interesting, nor are linear
  ## combinations

  Y <- ref(X) # row echelon form
  Y <- Y[!apply(Y == 0, 1, all), , drop=FALSE]
  pivot <- apply(Y != 0, 1, which.max)
  if (!setequal(1L:p, pivot))
      warning(sprintf("singular \'X\', dropping column(s) %s",
                      paste(setdiff(1L:p, pivot), collapse = ", ")),
              immediate. = TRUE)
  X <- X[, pivot, drop=FALSE]
  nv <- ncol(X)
  
  ## start of the algorithm, works with column names
  
  R <- cor(X)
  trR <- nv
  prv <- rep("", nv)
  scree <- rep(NA, nv)

  for (i in 1:nv) {
    h <- rowSums(R^2)
    prv[i] <- names(h)[j <- which.max(h)]
    vv <- R[j, j]
    if (vv < 1e-8) break
    R <- R[-j, -j, drop = FALSE] - crossprod(R[j, -j, drop = FALSE] / sqrt(vv))    
    scree[i] <- 1 - trace(R) / trR
  }

  names(scree) <- prv
  if (!as.names)
    prv <- match(prv, orignms)
  attr(prv, "scree") <- scree
  prv
}
